
# Dataclasses

New easy way of building classes since python37.

## Context

When writing a class that contains data, you often repeat yourself.

eg
```python
class Time:
    def __init__(self, hour: int, min: int = 0, sec: int = 0):
        self.hour: int = hour
        self.min: int = min
        self.sec: int = sec
```

Here, I wrote 3 time each attribute name !

The dataclass decorator helps you keeping your code shorter.

## Usage

### Simple usage

```python
from dataclasses import dataclass

@dataclass
class Time:
    hour: int
    min: int = 0
    sec: int = 0
```

It is way more compact, and easy to read and write !

You can also have a printing of the class (see [./00_time.py]) 



### Mutable default argument

To default the attribute, you can simply add a ` = default_value` at the end of its declaration.

**But**, you shouldn't use that for mutable arguments.

Instead, use the `field` function, with a `default_factory` argument.

eg at [./00_mutable.py]

```python
from dataclasses import dataclass, field
from typing import List

@dataclass
class Stock:
    products: List[str] = field(default_factory=list, init=False)

    def add_to_stock(self, product: str):
        self.products.append(product)
```

## Advanced usage

### The init variables

Sometimes, we don't want some parameters given to the constructor to end up as attribute of the final object.

For instance, let's say that we can instantiate our `Time` object with a timezone attribute. We want it to directly impact the hours, not keeping it as a permanent information.

For that, we have the `InitVar` type specifier, and the `__post_init__` function.

eg at [./00_time.py]

```python
from dataclasses import dataclass, InitVar
@dataclass
class Time:
    hour: int
    min: int = 0
    sec: int = 0

    time_zone: InitVar[int] = 0

    def __post_init__(self, time_zone: int):
        self.hour = (self.hour + time_zone) % 24
```

### Class variables

If you want a variable to belong to the class, and not to the objects, you can use the `ClassVar` specifier:

```python
from dataclasses import dataclass, InitVar
from typing import ClassVar

@dataclass
class Time:
    hour: int
    min: int = 0
    sec: int = 0

    time_zone: InitVar[int] = 0

    FORMAT: ClassVar[int] = 12  # 12 or 24 ?

    def __post_init__(self, time_zone: int):
        self.hour = (self.hour + time_zone) % self.FORMAT
```

### The field function

It allow you (among other things)

 - To say if the attribute should appear in the constructor (`init` parameter)
 - To default mutable attributes (through the `default_factory` argument`)

[Doc](https://docs.python.org/3/library/dataclasses.html#dataclasses.field)

## More information

See [python doc](https://docs.python.org/3/library/dataclasses.html)
